package fr.feufochmar.scribbly;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Scribbly command. A command contains three parts : an identifier, attributes, and child values.
 * The identifier is the only mandatory element.
 * 
 * @author Feufochmar
 *
 */
public class ScribblyCommand implements ScribblyElement
{
  /**
   * The identifier.
   */
  private String m_identifier = "";
  /**
   * The map of attributes.
   */
  private final Map<String, String> m_attributes = new HashMap<String, String>();
  /**
   * The list of child values.
   */
  private final List<ScribblyElement> m_values = new ArrayList<ScribblyElement>();
  
  /**
   * Constructor.
   */
  public ScribblyCommand()
  {
  }
  
  /**
   * Setter for the identifier.
   * @param identifier
   */
  public void setIdentifier(String identifier)
  {
    m_identifier = identifier;
  }
  
  /**
   * Getter for the identifier.
   * @return the identifier.
   */
  public String getIdentifier()
  {
    return m_identifier;
  }
  
  /**
   * Set the value of an attribute from a list of ScribblyText.
   * @param name Name of the attribute
   * @param value Value of the attribute
   */
  public void setAttribute(String name, List<ScribblyText> value)
  {
    String val = String.join("", value.stream().map(i -> i.getValue()).toList());
    m_attributes.put(name, val);
  }
  
  /**
   * Set the value of an attribute from a String.
   * @param name Name of the attribute.
   * @param value Value of the attribute.
   */
  public void setAttribute(String name, String value)
  {
    m_attributes.put(name, value);
  }
  
  /**
   * Get an attribute from its name.
   * @param name Name of the attribute
   * @return The value of the attribute, as a list of ScribblyText. Return an empty list if the attribute is not set.
   */
  public String getAttribute(String name)
  {
    return m_attributes.getOrDefault(name, "");
  }
  
  /**
   * Return wether the command has the given attribute.
   * @param name The attribute to check.
   * @return True if the command has the given attribute, false otherwise.
   */
  public boolean hasAttribute(String name)
  {
    return m_attributes.containsKey(name);
  }
  
  /**
   * Return the set of attributes.
   */
  public Set<Entry<String, String>> getAttributes()
  {
    return m_attributes.entrySet();
  }
  
  /**
   * Add a child value to the command.
   * @param value The child value to add.
   */
  public void addValue(ScribblyElement value)
  {
    m_values.add(value);
  }
  
  /**
   * Get the list of child values.
   * @return The list of child values.
   */
  public List<ScribblyElement> getValues()
  {
    return m_values;
  }

  /**
   * Implementation of the visitor pattern.
   */
  @Override
  public void visit(ScribblyVisitor visitor)
  {
    visitor.onBeginCommand(this);
    for (ScribblyElement elem: m_values)
    {
      elem.visit(visitor);
    }
    visitor.onEndCommand(this);
    
  }
}
