package fr.feufochmar.scribbly;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CharStreams;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.ErrorNode;
import org.antlr.v4.runtime.tree.TerminalNode;

import fr.feufochmar.scribbly.ScribblyParser.AttributeKeyContext;
import fr.feufochmar.scribbly.ScribblyParser.AttributeValueContext;
import fr.feufochmar.scribbly.ScribblyParser.CommandContext;
import fr.feufochmar.scribbly.ScribblyParser.CommandNameContext;
import fr.feufochmar.scribbly.ScribblyParser.CommandValContext;
import fr.feufochmar.scribbly.ScribblyParser.DocumentContext;
import fr.feufochmar.scribbly.ScribblyParser.ItemContext;
import fr.feufochmar.scribbly.ScribblyParser.TextContext;


/**
 * A scribbly document builder. Implements the ScribblyParserListener to construct the document content.
 * @author Feufochmar
 *
 */
public class ScribblyDocumentParser implements ScribblyParserListener
{
  private interface TerminalVisitor
  {
    void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node);
  }
  /**
   * Current context of visited terminal
   * @author Feufochmar
   */
  private enum Context implements TerminalVisitor
  {
    DOCUMENT
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        // Ignore terminals of document (can only be text or EOF)
      }
    },
    ITEM
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        // Ignore terminals of items (should be text or commands)
      }
    },
    COMMAND
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        // If terminal is ']', it's the end of an attribute => add the attribute and reset the temporary data
        if (node.getSymbol().getType() == ScribblyParser.RBLOCK)
        {
          // Set the attribute value
          docParser.m_currentCommand.peekFirst().setAttribute(docParser.m_currentAttributeKey, docParser.m_currentAttributeValue);
          // Reset the temporary data
          docParser.m_currentAttributeKey = "";
          docParser.m_currentAttributeValue.clear();
        }
        // Ignore other terminals of commands (they are part of the syntax of commands)
      }
    },
    COMMAND_NAME
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        docParser.m_currentCommand.peekFirst().setIdentifier(node.getText());
      }
    },
    COMMAND_ATTRIBUTE_KEY
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        docParser.m_currentAttributeKey = node.getText();
      }
    },
    COMMAND_ATTRIBUTE_VALUE
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        if (!docParser.m_isEscaping && node.getSymbol().getType() == ScribblyParser.ARROBE)
        {
          docParser.m_isEscaping = true;
        }
        else
        {
          docParser.m_currentAttributeValue.add(new ScribblyText(node.getText()));
          docParser.m_isEscaping = false;
        }
      }
    },
    COMMAND_VALUE
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        // Ignore terminals of command value (should be text or commands)
      }
    },
    TEXT
    {
      public void visitTerminal(ScribblyDocumentParser docParser, TerminalNode node)
      {
        if (!docParser.m_isEscaping && node.getSymbol().getType() == ScribblyParser.ARROBE)
        {
          docParser.m_isEscaping = true;
        }
        else
        {
          ScribblyText sText = new ScribblyText(node.getText());
          if (docParser.m_currentCommand.isEmpty())
          {
            docParser.m_document.addElement(sText);
          }
          else
          {
            docParser.m_currentCommand.peekFirst().addValue(sText);
          }
          docParser.m_isEscaping = false;
        }
      }
    };
  }
  private final Deque<Context> m_currentContext = new LinkedList<Context>();
  
  /**
   * Document built.
   */
  private final ScribblyDocument m_document = new ScribblyDocument();
  
  /**
   * Current command.
   */
  private final Deque<ScribblyCommand> m_currentCommand = new LinkedList<ScribblyCommand>();
  
  /**
   * Current attribute key.
   */
  private String m_currentAttributeKey = "";
  
  /**
   * Current attribute value.
   */
  private final List<ScribblyText> m_currentAttributeValue = new ArrayList<ScribblyText>();
  
  /**
   * Is escaping a character ?
   */
  private boolean m_isEscaping = false;
  
  /**
   * Constructor from a stream. Parse the input stream to initialize the document.
   * @param stream Input to parse
   * @throws IOException
   */
  public ScribblyDocumentParser(InputStream stream) throws IOException
  {
    parseInput(CharStreams.fromStream(stream));
  }
  
  /**
   * Constructor from a reader. Parse the input reader to initialize the document.
   * @param reader Reader to parse
   * @throws IOException
   */
  public ScribblyDocumentParser(Reader reader) throws IOException
  {
    parseInput(CharStreams.fromReader(reader));
  }
  
  /**
   * Constructor from a path to a file.
   * @param path File to parse
   * @throws FileNotFoundException
   * @throws IOException
   */
  public ScribblyDocumentParser(String fileName) throws FileNotFoundException, IOException
  {
    parseInput(CharStreams.fromFileName(fileName));
  }
  
  /**
   * Parse the input.
   * @param antlrStream The stream to parse.
   */
  private void parseInput(CharStream antlrStream)
  {
    ScribblyLexer lexer = new ScribblyLexer(antlrStream);
    ScribblyParser parser = new ScribblyParser(new CommonTokenStream(lexer));
    parser.addParseListener(this);
    parser.document();
  }
  
  /**
   * Return the parsed document.
   * @return
   */
  public ScribblyDocument document()
  {
    return m_document;
  }

  @Override
  public void visitTerminal(TerminalNode node)
  {
    // Current context shouldn't be empty
    if (m_currentContext.isEmpty())
    {
      return;
    }
    m_currentContext.peekFirst().visitTerminal(this, node);
  }

  @Override
  public void visitErrorNode(ErrorNode node)
  {
    // TODO Auto-generated method stub
    // Throw exception ?
  }

  @Override
  public void enterEveryRule(ParserRuleContext ctx)
  {
    // Nothing to do
  }

  @Override
  public void exitEveryRule(ParserRuleContext ctx)
  {
    // Nothing to do
  }

  @Override
  public void enterDocument(DocumentContext ctx)
  {
    m_currentContext.addFirst(Context.DOCUMENT);
  }

  @Override
  public void exitDocument(DocumentContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterItem(ItemContext ctx)
  {
    m_currentContext.addFirst(Context.ITEM);
  }

  @Override
  public void exitItem(ItemContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterCommand(CommandContext ctx)
  {
    m_currentContext.addFirst(Context.COMMAND);
    ScribblyCommand command = new ScribblyCommand();
    if (m_currentCommand.isEmpty())
    {
      m_document.addElement(command);
    }
    else
    {
      m_currentCommand.peekFirst().addValue(command);
    }
    m_currentCommand.addFirst(command);
  }

  @Override
  public void exitCommand(CommandContext ctx)
  {
    m_currentContext.removeFirst();
    m_currentCommand.removeFirst();
  }

  @Override
  public void enterCommandName(CommandNameContext ctx)
  {
    m_currentContext.addFirst(Context.COMMAND_NAME);
  }

  @Override
  public void exitCommandName(CommandNameContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterAttributeKey(AttributeKeyContext ctx)
  {
    m_currentContext.addFirst(Context.COMMAND_ATTRIBUTE_KEY);
  }

  @Override
  public void exitAttributeKey(AttributeKeyContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterAttributeValue(AttributeValueContext ctx)
  {
    m_currentContext.addFirst(Context.COMMAND_ATTRIBUTE_VALUE);
  }

  @Override
  public void exitAttributeValue(AttributeValueContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterCommandVal(CommandValContext ctx)
  {
    m_currentContext.addFirst(Context.COMMAND_VALUE);
  }

  @Override
  public void exitCommandVal(CommandValContext ctx)
  {
    m_currentContext.removeFirst();
  }

  @Override
  public void enterText(TextContext ctx)
  {
    m_currentContext.addFirst(Context.TEXT);
  }

  @Override
  public void exitText(TextContext ctx)
  {
    m_currentContext.removeFirst();
  }

}
